; frogger.asm

; OUT ports
#define PORT_CONTROL $E ; 1110 - bit 0 is low
#define PORT_LCD $D     ; 1101 - bit 1 is low

; IN ports
#define PORT_BUTTONS $E ; 1110 - bit 0 is low

; Bit flags in LCDCONTROL port
#define LCD_REG_COMMAND $4
#define LCD_REG_DATA $6

; LCD commands
#define LCD_COMMAND_INTERFACE8 $38  ; 8-bit interface, 2-line display, 5x8 font
#define LCD_COMMAND_INTERFACE4 $28  ; 4-bit interface, 2-line display, 5x8 font
#define LCD_COMMAND_DISPLAY $0C     ; display on, cursor off, blinking cursor off
#define LCD_COMMAND_CLEAR $01       ; clear display, home cursor
#define LCD_COMMAND_CGRAM_ADDR_0 $40 ; OR the desired CGRAM address with $40
#define LCD_COMMAND_CURSOR_POS_LINE_1 $80 ; OR the desired cursor position with $80 to create a cursor position command. Line 2 begins at pos 64
#define LCD_COMMAND_CURSOR_POS_LINE_2 $C0  
#define LCD_COMMAND_CURSOR_POS_GUESS $CC

; LCD timing constants
#define LCD_CLEAR_HOME_DELAY_US 1520
#define LCD_SINGLE_COMMAND_DELAY_US 37
#define CPU_CLOCKS_PER_US 2.4576

; memory locations
#define LCD_BUFFER_INDEX $00A
#define LCD_CONTROL_STATE $00E
#define RETURN_ADDR_0 $00F
#define RETURN_ADDR_1 $006
#define PREV_BUTTON_STATE $00D
#define NEW_BUTTON_STATE $00C
#define LCD_DELAY_1 $010
#define LCD_DELAY_2 $011
#define LCD_BUFFER $020
#define AUDIO_LOW_CONTROL_STATE $030
#define AUDIO_HIGH_CONTROL_STATE $031
#define TONE_PERIOD_0 $032
#define TONE_PERIOD_1 $033
#define TONE_PERIOD_2 $034
#define TONE_COUNTER_0 $038
#define TONE_COUNTER_1 $039
#define TONE_COUNTER_2 $03A
#define TONE_DURATION_0 $035
#define TONE_DURATION_1 $036
#define TONE_DURATION_2 $037
#define DEBOUNCE_COUNTER_0 $038
#define DEBOUNCE_COUNTER_1 $039
#define DEBOUNCE_COUNTER_2 $03A
#define DELAY_COUNTER_0 $038
#define DELAY_COUNTER_1 $039
#define DELAY_COUNTER_2 $03A
#define DELAY_COUNTER_3 $03B
#define BLINK_COUNTER $03C
#define FROG_X $03D
#define FROG_Y $03E
#define TEMP $03F
#define PLAYFIELD_INDEX_0 $000
#define PLAYFIELD_INDEX_1 $001
#define PLAYFIELD_DELTA $002
#define PLAYFIELD_NEW_VALUE $003
#define PLAYFIELD_RETURN $004
#define PLAYFIELD $040 ; 64 nibbles, $40 - $79
#define TONE_PERIOD_DELTA $005
#define COLLISION_RETURN $007
#define COLLISION_FLAG $008

; buttons
#define DEBOUNCE_TIME_0 $F
#define DEBOUNCE_TIME_1 $F
#define DEBOUNCE_TIME_2 $F
#define BUTTON_LEFT $1
#define BUTTON_NOT_LEFT $E
#define BUTTON_RIGHT $2
#define BUTTON_NOT_RIGHT $D
#define BUTTON_DOWN $4
#define BUTTON_NOT_DOWN $B
#define BUTTON_UP $8
#define BUTTON_NOT_UP $7

init:
    ; To initialize the LCD from an unknown state, we must first set it to 8-bit mode three times. Then it can be set to 4-bit mode.

    ; prepare to send an LCD command
    lit #LCD_REG_COMMAND
    st LCD_CONTROL_STATE
    ; remember where to return 
    lit #0
    st RETURN_ADDR_1
    lit #0
    st RETURN_ADDR_0
    ; zero more nibbles to read from the buffer
    lit #0
    st LCD_BUFFER_INDEX
    ; the command is INTERFACE8
    lit #<LCD_COMMAND_INTERFACE8
    jmp lcd_write_nibble

next0:
    ; remember where to return
    lit #0
    st RETURN_ADDR_0
    jmp lcd_long_delay

delaynext0:
    ; remember where to return
    lit #0
    st RETURN_ADDR_1
    lit #1
    st RETURN_ADDR_0
    ; zero more nibbles to read from the buffer
    lit #0
    st LCD_BUFFER_INDEX
    ; the command is INTERFACE8
    lit #<LCD_COMMAND_INTERFACE8
    jmp lcd_write_nibble

next1:
    ; remember where to return
    lit #1
    st RETURN_ADDR_0
    jmp lcd_long_delay

delaynext1:
    ; remember where to return
    lit #0
    st RETURN_ADDR_1
    lit #2
    st RETURN_ADDR_0
    ; zero more nibbles to read from the buffer
    lit #0
    st LCD_BUFFER_INDEX
    ; the command is INTERFACE8
    lit #<LCD_COMMAND_INTERFACE8
    jmp lcd_write_nibble

next2:
    ; remember where to return
    lit #0
    st RETURN_ADDR_1
    lit #3
    st RETURN_ADDR_0
    ; zero more nibbles to read from the buffer
    lit #0
    st LCD_BUFFER_INDEX
    ; the command is INTERFACE4
    lit #<LCD_COMMAND_INTERFACE4
    jmp lcd_write_nibble

    ; The LCD is now in 4-bit mode, and we can send byte-wide commands as a pair of nibble, high nibble first.

next3:
    ; prepare to send an LCD command
    lit #LCD_REG_COMMAND
    st LCD_CONTROL_STATE

    lit #<LCD_COMMAND_INTERFACE4
    st LCD_BUFFER+5
    lit #>LCD_COMMAND_INTERFACE4
    st LCD_BUFFER+4
    lit #<LCD_COMMAND_DISPLAY
    st LCD_BUFFER+3
    lit #>LCD_COMMAND_DISPLAY
    st LCD_BUFFER+2
    lit #<LCD_COMMAND_CLEAR
    st LCD_BUFFER+1
    lit #>LCD_COMMAND_CLEAR
    st LCD_BUFFER+0

    ; remember where to return
    lit #0
    st RETURN_ADDR_1
    lit #4
    st RETURN_ADDR_0
    ; data begins at buffer index 5
    lit #5
    jmp lcd_write_buffer

next4:
    ; remember where to return
    lit #2
    st RETURN_ADDR_0
    jmp lcd_long_delay

delaynext2:
    lit #<LCD_COMMAND_CGRAM_ADDR_0
    st LCD_BUFFER+1
    lit #>LCD_COMMAND_CGRAM_ADDR_0
    st LCD_BUFFER+0

    ; remember where to return
    lit #0
    st RETURN_ADDR_1
    lit #5
    st RETURN_ADDR_0
    ; data begins at buffer index 1
    lit #1
    jmp lcd_write_buffer

next5:
    ; prepare to send LCD data
    lit #LCD_REG_DATA
    st LCD_CONTROL_STATE

    ; 0 - empty
    lit #$0
    st LCD_BUFFER+15
    lit #$0
    st LCD_BUFFER+14
    lit #$0
    st LCD_BUFFER+13
    lit #$0
    st LCD_BUFFER+12
    lit #$0
    st LCD_BUFFER+11
    lit #$0
    st LCD_BUFFER+10
    lit #$0
    st LCD_BUFFER+9
    lit #$0
    st LCD_BUFFER+8
    lit #$0
    st LCD_BUFFER+7
    lit #$0
    st LCD_BUFFER+6
    lit #$0
    st LCD_BUFFER+5
    lit #$0
    st LCD_BUFFER+4
    lit #$0
    st LCD_BUFFER+3
    lit #$0
    st LCD_BUFFER+2
    lit #$0
    st LCD_BUFFER+1
    lit #$0
    st LCD_BUFFER+0

    ; remember where to return
    lit #0
    st RETURN_ADDR_1
    lit #6
    st RETURN_ADDR_0
    ; data begins at buffer index 15
    lit #15
    jmp lcd_write_buffer

next6:
    ; 1 - lower log
    lit #$0
    st LCD_BUFFER+15
    lit #$0
    st LCD_BUFFER+14
    lit #$0
    st LCD_BUFFER+13
    lit #$0
    st LCD_BUFFER+12
    lit #$0
    st LCD_BUFFER+11
    lit #$0
    st LCD_BUFFER+10
    lit #$0
    st LCD_BUFFER+9
    lit #$0
    st LCD_BUFFER+8
    lit #$0
    st LCD_BUFFER+7
    lit #$0
    st LCD_BUFFER+6
    lit #$0
    st LCD_BUFFER+5
    lit #$E
    st LCD_BUFFER+4
    lit #$1
    st LCD_BUFFER+3
    lit #$1
    st LCD_BUFFER+2
    lit #$0
    st LCD_BUFFER+1
    lit #$E
    st LCD_BUFFER+0

    ; remember where to return
    lit #0
    st RETURN_ADDR_1
    lit #7
    st RETURN_ADDR_0
    ; data begins at buffer index 15
    lit #15
    jmp lcd_write_buffer

next7:
    ; 2 - lower frog
    lit #$0
    st LCD_BUFFER+15
    lit #$0
    st LCD_BUFFER+14
    lit #$0
    st LCD_BUFFER+13
    lit #$0
    st LCD_BUFFER+12
    lit #$0
    st LCD_BUFFER+11
    lit #$0
    st LCD_BUFFER+10
    lit #$0
    st LCD_BUFFER+9
    lit #$0
    st LCD_BUFFER+8
    lit #$0
    st LCD_BUFFER+7
    lit #$E
    st LCD_BUFFER+6
    lit #$1
    st LCD_BUFFER+5
    lit #$5
    st LCD_BUFFER+4
    lit #$0
    st LCD_BUFFER+3
    lit #$E
    st LCD_BUFFER+2
    lit #$1
    st LCD_BUFFER+1
    lit #$B
    st LCD_BUFFER+0

    ; remember where to return
    lit #0
    st RETURN_ADDR_1
    lit #8
    st RETURN_ADDR_0
    ; data begins at buffer index 15
    lit #15
    jmp lcd_write_buffer

next8:
    ; 3 - upper log
    lit #$0
    st LCD_BUFFER+15
    lit #$0
    st LCD_BUFFER+14
    lit #$0
    st LCD_BUFFER+13
    lit #$E
    st LCD_BUFFER+12
    lit #$0
    st LCD_BUFFER+11
    lit #$E
    st LCD_BUFFER+10
    lit #$0
    st LCD_BUFFER+9
    lit #$0
    st LCD_BUFFER+8
    lit #$0
    st LCD_BUFFER+7
    lit #$0
    st LCD_BUFFER+6
    lit #$0
    st LCD_BUFFER+5
    lit #$0
    st LCD_BUFFER+4
    lit #$0
    st LCD_BUFFER+3
    lit #$0
    st LCD_BUFFER+2
    lit #$0
    st LCD_BUFFER+1
    lit #$0
    st LCD_BUFFER+0

    ; remember where to return
    lit #0
    st RETURN_ADDR_1
    lit #9
    st RETURN_ADDR_0
    ; data begins at buffer index 15
    lit #15
    jmp lcd_write_buffer

next9:
    ; 4 - two logs
    lit #$0
    st LCD_BUFFER+15
    lit #$0
    st LCD_BUFFER+14
    lit #$0
    st LCD_BUFFER+13
    lit #$E
    st LCD_BUFFER+12
    lit #$0
    st LCD_BUFFER+11
    lit #$E
    st LCD_BUFFER+10
    lit #$0
    st LCD_BUFFER+9
    lit #$0
    st LCD_BUFFER+8
    lit #$0
    st LCD_BUFFER+7
    lit #$0
    st LCD_BUFFER+6
    lit #$0
    st LCD_BUFFER+5
    lit #$E
    st LCD_BUFFER+4
    lit #$1
    st LCD_BUFFER+3
    lit #$1
    st LCD_BUFFER+2
    lit #$0
    st LCD_BUFFER+1
    lit #$E
    st LCD_BUFFER+0

    ; remember where to return
    lit #0
    st RETURN_ADDR_1
    lit #10
    st RETURN_ADDR_0
    ; data begins at buffer index 15
    lit #15
    jmp lcd_write_buffer

next10:
    ; 5 - log above frog
    lit #$0
    st LCD_BUFFER+15
    lit #$0
    st LCD_BUFFER+14
    lit #$0
    st LCD_BUFFER+13
    lit #$E
    st LCD_BUFFER+12
    lit #$0
    st LCD_BUFFER+11
    lit #$E
    st LCD_BUFFER+10
    lit #$0
    st LCD_BUFFER+9
    lit #$0
    st LCD_BUFFER+8
    lit #$0
    st LCD_BUFFER+7
    lit #$E
    st LCD_BUFFER+6
    lit #$1
    st LCD_BUFFER+5
    lit #$5
    st LCD_BUFFER+4
    lit #$0
    st LCD_BUFFER+3
    lit #$E
    st LCD_BUFFER+2
    lit #$1
    st LCD_BUFFER+1
    lit #$B
    st LCD_BUFFER+0

    ; remember where to return
    lit #0
    st RETURN_ADDR_1
    lit #11
    st RETURN_ADDR_0
    ; data begins at buffer index 15
    lit #15
    jmp lcd_write_buffer

next11:
    ; 6 - upper frog
    lit #$0
    st LCD_BUFFER+15
    lit #$E
    st LCD_BUFFER+14
    lit #$1
    st LCD_BUFFER+13
    lit #$5
    st LCD_BUFFER+12
    lit #$0
    st LCD_BUFFER+11
    lit #$E
    st LCD_BUFFER+10
    lit #$1
    st LCD_BUFFER+9
    lit #$B
    st LCD_BUFFER+8
    lit #$0
    st LCD_BUFFER+7
    lit #$0
    st LCD_BUFFER+6
    lit #$0
    st LCD_BUFFER+5
    lit #$0
    st LCD_BUFFER+4
    lit #$0
    st LCD_BUFFER+3
    lit #$0
    st LCD_BUFFER+2
    lit #$0
    st LCD_BUFFER+1
    lit #$0
    st LCD_BUFFER+0

    ; remember where to return
    lit #0
    st RETURN_ADDR_1
    lit #12
    st RETURN_ADDR_0
    ; data begins at buffer index 15
    lit #15
    jmp lcd_write_buffer

next12:
    ; 7 - frog above log
    lit #$0
    st LCD_BUFFER+15
    lit #$E
    st LCD_BUFFER+14
    lit #$1
    st LCD_BUFFER+13
    lit #$5
    st LCD_BUFFER+12
    lit #$0
    st LCD_BUFFER+11
    lit #$E
    st LCD_BUFFER+10
    lit #$1
    st LCD_BUFFER+9
    lit #$B
    st LCD_BUFFER+8
    lit #$0
    st LCD_BUFFER+7
    lit #$0
    st LCD_BUFFER+6
    lit #$0
    st LCD_BUFFER+5
    lit #$E
    st LCD_BUFFER+4
    lit #$1
    st LCD_BUFFER+3
    lit #$1
    st LCD_BUFFER+2
    lit #$0
    st LCD_BUFFER+1
    lit #$E
    st LCD_BUFFER+0

    ; remember where to return
    lit #0
    st RETURN_ADDR_1
    lit #13
    st RETURN_ADDR_0
    ; data begins at buffer index 15
    lit #15
    jmp lcd_write_buffer

next13:
    jmp init_playfield

init_playfield:
    lit #0
    st PLAYFIELD
    st PLAYFIELD+1
    st PLAYFIELD+3
    st PLAYFIELD+4
    st PLAYFIELD+5
    st PLAYFIELD+7
    st PLAYFIELD+10
    st PLAYFIELD+9
    st PLAYFIELD+11
    st PLAYFIELD+12
    st PLAYFIELD+13
    st PLAYFIELD+15

    st PLAYFIELD+17
    st PLAYFIELD+18
    st PLAYFIELD+19
    st PLAYFIELD+21
    st PLAYFIELD+23
    st PLAYFIELD+25
    st PLAYFIELD+26
    st PLAYFIELD+27
    st PLAYFIELD+29
    st PLAYFIELD+30
    st PLAYFIELD+31

    st PLAYFIELD+32
    st PLAYFIELD+33
    st PLAYFIELD+34
    st PLAYFIELD+36
    st PLAYFIELD+37
    st PLAYFIELD+38
    st PLAYFIELD+40
    st PLAYFIELD+41
    st PLAYFIELD+43
    st PLAYFIELD+44
    st PLAYFIELD+45
    st PLAYFIELD+46

    st PLAYFIELD+48
    st PLAYFIELD+50
    st PLAYFIELD+51
    st PLAYFIELD+52
    st PLAYFIELD+54
    st PLAYFIELD+55
    st PLAYFIELD+56
    st PLAYFIELD+57
    st PLAYFIELD+58
    st PLAYFIELD+59
    st PLAYFIELD+60
    st PLAYFIELD+62
    st PLAYFIELD+63

    lit #1
    st PLAYFIELD+2
    st PLAYFIELD+6
    st PLAYFIELD+8
    st PLAYFIELD+10
    st PLAYFIELD+14

    st PLAYFIELD+16
    st PLAYFIELD+20
    st PLAYFIELD+22
    st PLAYFIELD+24
    st PLAYFIELD+28

    st PLAYFIELD+35
    st PLAYFIELD+39
    st PLAYFIELD+42
    st PLAYFIELD+47

    st PLAYFIELD+49
    st PLAYFIELD+53
    st PLAYFIELD+61

    ; frog
    lit #0
    st PLAYFIELD+54

    lit #3
    st FROG_Y
    lit #6
    st FROG_X
    lit #0
    st COLLISION_FLAG

    ; remember where to return
    lit #0
    st PLAYFIELD_RETURN

    jmp draw_playfield

playfieldnext0:
    jmp input

draw_playfield:
    ld COLLISION_FLAG
    jnz no_collision
    ; add frog
    ld FROG_X
    st PLAYFIELD_INDEX_0
    ld FROG_Y
    st PLAYFIELD_INDEX_1
    lit #2
    st PLAYFIELD_DELTA
    ; remember where to return
    lit #1
    st RETURN_ADDR_1
    lit #11
    st RETURN_ADDR_0
    jmp modify_playfield_value

next27:
    ; collision check
    ld PLAYFIELD_NEW_VALUE
    cmpi #2
    jz no_collision

    ; collision!
    lit #1
    st COLLISION_FLAG

    ; subtract frog
    lit #-2
    st PLAYFIELD_DELTA
    ; remember where to return
    lit #0
    st RETURN_ADDR_1
    lit #14
    st RETURN_ADDR_0
    jmp modify_playfield_value

next14:       
no_collision:
    ; prepare to send an LCD command
    lit #LCD_REG_COMMAND
    st LCD_CONTROL_STATE

    lit #<LCD_COMMAND_CURSOR_POS_LINE_1
    st LCD_BUFFER+1
    lit #>LCD_COMMAND_CURSOR_POS_LINE_1
    st LCD_BUFFER+0

    ; remember where to return
    lit #1
    st RETURN_ADDR_1
    lit #0
    st RETURN_ADDR_0
    ; data begins at buffer index 1
    lit #1
    jmp lcd_write_buffer

next16:
    ; the high nibble of the text buffer is always 0
    lit #0
    st LCD_BUFFER+15
    lit #0
    st LCD_BUFFER+13
    lit #0
    st LCD_BUFFER+11
    lit #0
    st LCD_BUFFER+9
    lit #0
    st LCD_BUFFER+7
    lit #0
    st LCD_BUFFER+5
    lit #0
    st LCD_BUFFER+3
    lit #0
    st LCD_BUFFER+1

    ; prepare to send LCD data
    lit #LCD_REG_DATA
    st LCD_CONTROL_STATE

    ; row 0, left
    ld PLAYFIELD
    addm PLAYFIELD
    addm PLAYFIELD
    addm PLAYFIELD+16
    st LCD_BUFFER+14
    ld PLAYFIELD+1
    addm PLAYFIELD+1
    addm PLAYFIELD+1
    addm PLAYFIELD+17
    st LCD_BUFFER+12
    ld PLAYFIELD+2
    addm PLAYFIELD+2
    addm PLAYFIELD+2
    addm PLAYFIELD+18
    st LCD_BUFFER+10
    ld PLAYFIELD+3
    addm PLAYFIELD+3
    addm PLAYFIELD+3
    addm PLAYFIELD+19
    st LCD_BUFFER+8
    ld PLAYFIELD+4
    addm PLAYFIELD+4
    addm PLAYFIELD+4
    addm PLAYFIELD+20
    st LCD_BUFFER+6
    ld PLAYFIELD+5
    addm PLAYFIELD+5
    addm PLAYFIELD+5
    addm PLAYFIELD+21
    st LCD_BUFFER+4
    ld PLAYFIELD+6
    addm PLAYFIELD+6
    addm PLAYFIELD+6
    addm PLAYFIELD+22
    st LCD_BUFFER+2
    ld PLAYFIELD+7
    addm PLAYFIELD+7
    addm PLAYFIELD+7
    addm PLAYFIELD+23
    st LCD_BUFFER+0

    ; remember where to return
    lit #1
    st RETURN_ADDR_1
    lit #1
    st RETURN_ADDR_0
    ; data begins at buffer index 15
    lit #15
    jmp lcd_write_buffer

next17:
    ; row 0, right
    ld PLAYFIELD+8
    addm PLAYFIELD+8
    addm PLAYFIELD+8
    addm PLAYFIELD+24
    st LCD_BUFFER+14
    ld PLAYFIELD+9
    addm PLAYFIELD+9
    addm PLAYFIELD+9
    addm PLAYFIELD+25
    st LCD_BUFFER+12
    ld PLAYFIELD+10
    addm PLAYFIELD+10
    addm PLAYFIELD+10
    addm PLAYFIELD+26
    st LCD_BUFFER+10
    ld PLAYFIELD+11
    addm PLAYFIELD+11
    addm PLAYFIELD+11
    addm PLAYFIELD+27
    st LCD_BUFFER+8
    ld PLAYFIELD+12
    addm PLAYFIELD+12
    addm PLAYFIELD+12
    addm PLAYFIELD+28
    st LCD_BUFFER+6
    ld PLAYFIELD+13
    addm PLAYFIELD+13
    addm PLAYFIELD+13
    addm PLAYFIELD+29
    st LCD_BUFFER+4
    ld PLAYFIELD+14
    addm PLAYFIELD+14
    addm PLAYFIELD+14
    addm PLAYFIELD+30
    st LCD_BUFFER+2
    ld PLAYFIELD+15
    addm PLAYFIELD+15
    addm PLAYFIELD+15
    addm PLAYFIELD+31
    st LCD_BUFFER+0

    ; remember where to return
    lit #1
    st RETURN_ADDR_1
    lit #2
    st RETURN_ADDR_0
    ; data begins at buffer index 15
    lit #15
    jmp lcd_write_buffer

next18:
    ; prepare to send an LCD command
    lit #LCD_REG_COMMAND
    st LCD_CONTROL_STATE

    lit #<LCD_COMMAND_CURSOR_POS_LINE_2
    st LCD_BUFFER+1
    lit #>LCD_COMMAND_CURSOR_POS_LINE_2
    st LCD_BUFFER+0

    ; remember where to return
    lit #1
    st RETURN_ADDR_1
    lit #3
    st RETURN_ADDR_0
    ; data begins at buffer index 1
    lit #1
    jmp lcd_write_buffer

next19:
    ; prepare to send LCD data
    lit #LCD_REG_DATA
    st LCD_CONTROL_STATE

    lit #0
    st LCD_BUFFER+1

    ; row 1, left
    ld PLAYFIELD+32
    addm PLAYFIELD+32
    addm PLAYFIELD+32
    addm PLAYFIELD+48
    st LCD_BUFFER+14
    ld PLAYFIELD+33
    addm PLAYFIELD+33
    addm PLAYFIELD+33
    addm PLAYFIELD+49
    st LCD_BUFFER+12
    ld PLAYFIELD+34
    addm PLAYFIELD+34
    addm PLAYFIELD+34
    addm PLAYFIELD+50
    st LCD_BUFFER+10
    ld PLAYFIELD+35
    addm PLAYFIELD+35
    addm PLAYFIELD+35
    addm PLAYFIELD+51
    st LCD_BUFFER+8
    ld PLAYFIELD+36
    addm PLAYFIELD+36
    addm PLAYFIELD+36
    addm PLAYFIELD+52
    st LCD_BUFFER+6
    ld PLAYFIELD+37
    addm PLAYFIELD+37
    addm PLAYFIELD+37
    addm PLAYFIELD+53
    st LCD_BUFFER+4
    ld PLAYFIELD+38
    addm PLAYFIELD+38
    addm PLAYFIELD+38
    addm PLAYFIELD+54
    st LCD_BUFFER+2
    ld PLAYFIELD+39
    addm PLAYFIELD+39
    addm PLAYFIELD+39
    addm PLAYFIELD+55
    st LCD_BUFFER+0

    ; remember where to return
    lit #1
    st RETURN_ADDR_1
    lit #4
    st RETURN_ADDR_0
    ; data begins at buffer index 15
    lit #15
    jmp lcd_write_buffer

next20:
    ; row 1, right
    ld PLAYFIELD+40
    addm PLAYFIELD+40
    addm PLAYFIELD+40
    addm PLAYFIELD+56
    st LCD_BUFFER+14
    ld PLAYFIELD+41
    addm PLAYFIELD+41
    addm PLAYFIELD+41
    addm PLAYFIELD+57
    st LCD_BUFFER+12
    ld PLAYFIELD+42
    addm PLAYFIELD+42
    addm PLAYFIELD+42
    addm PLAYFIELD+58
    st LCD_BUFFER+10
    ld PLAYFIELD+43
    addm PLAYFIELD+43
    addm PLAYFIELD+43
    addm PLAYFIELD+59
    st LCD_BUFFER+8
    ld PLAYFIELD+44
    addm PLAYFIELD+44
    addm PLAYFIELD+44
    addm PLAYFIELD+60
    st LCD_BUFFER+6
    ld PLAYFIELD+45
    addm PLAYFIELD+45
    addm PLAYFIELD+45
    addm PLAYFIELD+61
    st LCD_BUFFER+4
    ld PLAYFIELD+46
    addm PLAYFIELD+46
    addm PLAYFIELD+46
    addm PLAYFIELD+62
    st LCD_BUFFER+2
    ld PLAYFIELD+47
    addm PLAYFIELD+47
    addm PLAYFIELD+47
    addm PLAYFIELD+63
    st LCD_BUFFER+0

    ; remember where to return
    lit #1
    st RETURN_ADDR_1
    lit #5
    st RETURN_ADDR_0
    ; data begins at buffer index 15
    lit #15
    jmp lcd_write_buffer

next21:
    ld COLLISION_FLAG
    jnz frog_erased

    ld FROG_X
    st PLAYFIELD_INDEX_0
    ld FROG_Y
    st PLAYFIELD_INDEX_1
    lit #-2
    st PLAYFIELD_DELTA
    ; remember where to return
    lit #1
    st RETURN_ADDR_1
    lit #12
    st RETURN_ADDR_0
    jmp modify_playfield_value

next28:
frog_erased:
    ld PLAYFIELD_RETURN
    jz playfieldnext0
    cmpi #1
    jz playfieldnext1
    cmpi #2
    jz playfieldnext2
    cmpi #3
    jz playfieldnext3
    cmpi #4
    jz playfieldnext4

    jmp halt

scroll_left:
    ld PLAYFIELD+16
    st TEMP
    ld PLAYFIELD+17
    st PLAYFIELD+16
    ld PLAYFIELD+18
    st PLAYFIELD+17
    ld PLAYFIELD+19
    st PLAYFIELD+18
    ld PLAYFIELD+20
    st PLAYFIELD+19
    ld PLAYFIELD+21
    st PLAYFIELD+20
    ld PLAYFIELD+22
    st PLAYFIELD+21
    ld PLAYFIELD+23
    st PLAYFIELD+22
    ld PLAYFIELD+24
    st PLAYFIELD+23
    ld PLAYFIELD+25
    st PLAYFIELD+24
    ld PLAYFIELD+26
    st PLAYFIELD+25
    ld PLAYFIELD+27
    st PLAYFIELD+26
    ld PLAYFIELD+28
    st PLAYFIELD+27
    ld PLAYFIELD+29
    st PLAYFIELD+28
    ld PLAYFIELD+30
    st PLAYFIELD+29
    ld PLAYFIELD+31
    st PLAYFIELD+30
    ld TEMP
    st PLAYFIELD+31

    ld PLAYFIELD+48
    st TEMP
    ld PLAYFIELD+49
    st PLAYFIELD+48
    ld PLAYFIELD+50
    st PLAYFIELD+49
    ld PLAYFIELD+51
    st PLAYFIELD+50
    ld PLAYFIELD+52
    st PLAYFIELD+51
    ld PLAYFIELD+53
    st PLAYFIELD+52
    ld PLAYFIELD+54
    st PLAYFIELD+53
    ld PLAYFIELD+55
    st PLAYFIELD+54
    ld PLAYFIELD+56
    st PLAYFIELD+55
    ld PLAYFIELD+57
    st PLAYFIELD+56
    ld PLAYFIELD+58
    st PLAYFIELD+57
    ld PLAYFIELD+59
    st PLAYFIELD+58
    ld PLAYFIELD+60
    st PLAYFIELD+59
    ld PLAYFIELD+61
    st PLAYFIELD+60
    ld PLAYFIELD+62
    st PLAYFIELD+61
    ld PLAYFIELD+63
    st PLAYFIELD+62
    ld TEMP
    st PLAYFIELD+63

    jmp return

scroll_right:
    ld PLAYFIELD+15
    st TEMP
    ld PLAYFIELD+14
    st PLAYFIELD+15
    ld PLAYFIELD+13
    st PLAYFIELD+14
    ld PLAYFIELD+12
    st PLAYFIELD+13
    ld PLAYFIELD+11
    st PLAYFIELD+12
    ld PLAYFIELD+10
    st PLAYFIELD+11
    ld PLAYFIELD+9
    st PLAYFIELD+10
    ld PLAYFIELD+8
    st PLAYFIELD+9
    ld PLAYFIELD+7
    st PLAYFIELD+8
    ld PLAYFIELD+6
    st PLAYFIELD+7
    ld PLAYFIELD+5
    st PLAYFIELD+6
    ld PLAYFIELD+4
    st PLAYFIELD+5
    ld PLAYFIELD+3
    st PLAYFIELD+4
    ld PLAYFIELD+2
    st PLAYFIELD+3
    ld PLAYFIELD+1
    st PLAYFIELD+2
    ld PLAYFIELD+0
    st PLAYFIELD+1
    ld TEMP
    st PLAYFIELD+0

    ld PLAYFIELD+47
    st TEMP
    ld PLAYFIELD+46
    st PLAYFIELD+47
    ld PLAYFIELD+45
    st PLAYFIELD+46
    ld PLAYFIELD+44
    st PLAYFIELD+45
    ld PLAYFIELD+43
    st PLAYFIELD+44
    ld PLAYFIELD+42
    st PLAYFIELD+43
    ld PLAYFIELD+41
    st PLAYFIELD+42
    ld PLAYFIELD+40
    st PLAYFIELD+41
    ld PLAYFIELD+39
    st PLAYFIELD+40
    ld PLAYFIELD+38
    st PLAYFIELD+39
    ld PLAYFIELD+37
    st PLAYFIELD+38
    ld PLAYFIELD+36
    st PLAYFIELD+37
    ld PLAYFIELD+35
    st PLAYFIELD+36
    ld PLAYFIELD+34
    st PLAYFIELD+35
    ld PLAYFIELD+33
    st PLAYFIELD+34
    ld PLAYFIELD+32
    st PLAYFIELD+33
    ld TEMP
    st PLAYFIELD+32

    jmp return

modify_playfield_value:
    ld PLAYFIELD_INDEX_1
    jnz ++
    ld PLAYFIELD_INDEX_0
    jnz +
    ld PLAYFIELD+0
    addm PLAYFIELD_DELTA
    st PLAYFIELD+0
    jmp modify_done
+   cmpi #1
    jnz +
    ld PLAYFIELD+1
    addm PLAYFIELD_DELTA
    st PLAYFIELD+1
    jmp modify_done
+   cmpi #2
    jnz +
    ld PLAYFIELD+2
    addm PLAYFIELD_DELTA
    st PLAYFIELD+2
    jmp modify_done
+   cmpi #3
    jnz +
    ld PLAYFIELD+3
    addm PLAYFIELD_DELTA
    st PLAYFIELD+3
    jmp modify_done
+   cmpi #4
    jnz +
    ld PLAYFIELD+4
    addm PLAYFIELD_DELTA
    st PLAYFIELD+4
    jmp modify_done
+   cmpi #5
    jnz +
    ld PLAYFIELD+5
    addm PLAYFIELD_DELTA
    st PLAYFIELD+5
    jmp modify_done
+   cmpi #6
    jnz +
    ld PLAYFIELD+6
    addm PLAYFIELD_DELTA
    st PLAYFIELD+6
    jmp modify_done
+   cmpi #7
    jnz +
    ld PLAYFIELD+7
    addm PLAYFIELD_DELTA
    st PLAYFIELD+7
    jmp modify_done
+   cmpi #8
    jnz +
    ld PLAYFIELD+8
    addm PLAYFIELD_DELTA
    st PLAYFIELD+8
    jmp modify_done
+   cmpi #9
    jnz +
    ld PLAYFIELD+9
    addm PLAYFIELD_DELTA
    st PLAYFIELD+9
    jmp modify_done
+   cmpi #10
    jnz +
    ld PLAYFIELD+10
    addm PLAYFIELD_DELTA
    st PLAYFIELD+10
    jmp modify_done
+   cmpi #11
    jnz +
    ld PLAYFIELD+11
    addm PLAYFIELD_DELTA
    st PLAYFIELD+11
    jmp modify_done
+   cmpi #12
    jnz +
    ld PLAYFIELD+12
    addm PLAYFIELD_DELTA
    st PLAYFIELD+12
    jmp modify_done
+   cmpi #13
    jnz +
    ld PLAYFIELD+13
    addm PLAYFIELD_DELTA
    st PLAYFIELD+13
    jmp modify_done
+   cmpi #14
    jnz +
    ld PLAYFIELD+14
    addm PLAYFIELD_DELTA
    st PLAYFIELD+14
    jmp modify_done
+   ld PLAYFIELD+15
    addm PLAYFIELD_DELTA
    st PLAYFIELD+15
    jmp modify_done

++  cmpi #1
    jnz ++
    ld PLAYFIELD_INDEX_0
    jnz +
    ld PLAYFIELD+16
    addm PLAYFIELD_DELTA
    st PLAYFIELD+16
    jmp modify_done
+   cmpi #1
    jnz +
    ld PLAYFIELD+17
    addm PLAYFIELD_DELTA
    st PLAYFIELD+17
    jmp modify_done
+   cmpi #2
    jnz +
    ld PLAYFIELD+18
    addm PLAYFIELD_DELTA
    st PLAYFIELD+18
    jmp modify_done
+   cmpi #3
    jnz +
    ld PLAYFIELD+19
    addm PLAYFIELD_DELTA
    st PLAYFIELD+19
    jmp modify_done
+   cmpi #4
    jnz +
    ld PLAYFIELD+20
    addm PLAYFIELD_DELTA
    st PLAYFIELD+20
    jmp modify_done
+   cmpi #5
    jnz +
    ld PLAYFIELD+21
    addm PLAYFIELD_DELTA
    st PLAYFIELD+21
    jmp modify_done
+   cmpi #6
    jnz +
    ld PLAYFIELD+22
    addm PLAYFIELD_DELTA
    st PLAYFIELD+22
    jmp modify_done
+   cmpi #7
    jnz +
    ld PLAYFIELD+23
    addm PLAYFIELD_DELTA
    st PLAYFIELD+23
    jmp modify_done
+   cmpi #8
    jnz +
    ld PLAYFIELD+24
    addm PLAYFIELD_DELTA
    st PLAYFIELD+24
    jmp modify_done
+   cmpi #9
    jnz +
    ld PLAYFIELD+25
    addm PLAYFIELD_DELTA
    st PLAYFIELD+25
    jmp modify_done
+   cmpi #10
    jnz +
    ld PLAYFIELD+26
    addm PLAYFIELD_DELTA
    st PLAYFIELD+26
    jmp modify_done
+   cmpi #11
    jnz +
    ld PLAYFIELD+27
    addm PLAYFIELD_DELTA
    st PLAYFIELD+27
    jmp modify_done
+   cmpi #12
    jnz +
    ld PLAYFIELD+28
    addm PLAYFIELD_DELTA
    st PLAYFIELD+28
    jmp modify_done
+   cmpi #13
    jnz +
    ld PLAYFIELD+29
    addm PLAYFIELD_DELTA
    st PLAYFIELD+29
    jmp modify_done
+   cmpi #14
    jnz +
    ld PLAYFIELD+30
    addm PLAYFIELD_DELTA
    st PLAYFIELD+30
    jmp modify_done
+   ld PLAYFIELD+31
    addm PLAYFIELD_DELTA
    st PLAYFIELD+31
    jmp modify_done

++  cmpi #2
    jnz ++
    ld PLAYFIELD_INDEX_0
    jnz +
    ld PLAYFIELD+32
    addm PLAYFIELD_DELTA
    st PLAYFIELD+32
    jmp modify_done
+   cmpi #1
    jnz +
    ld PLAYFIELD+33
    addm PLAYFIELD_DELTA
    st PLAYFIELD+33
    jmp modify_done
+   cmpi #2
    jnz +
    ld PLAYFIELD+34
    addm PLAYFIELD_DELTA
    st PLAYFIELD+34
    jmp modify_done
+   cmpi #3
    jnz +
    ld PLAYFIELD+35
    addm PLAYFIELD_DELTA
    st PLAYFIELD+35
    jmp modify_done
+   cmpi #4
    jnz +
    ld PLAYFIELD+36
    addm PLAYFIELD_DELTA
    st PLAYFIELD+36
    jmp modify_done
+   cmpi #5
    jnz +
    ld PLAYFIELD+37
    addm PLAYFIELD_DELTA
    st PLAYFIELD+37
    jmp modify_done
+   cmpi #6
    jnz +
    ld PLAYFIELD+38
    addm PLAYFIELD_DELTA
    st PLAYFIELD+38
    jmp modify_done
+   cmpi #7
    jnz +
    ld PLAYFIELD+39
    addm PLAYFIELD_DELTA
    st PLAYFIELD+39
    jmp modify_done
+   cmpi #8
    jnz +
    ld PLAYFIELD+40
    addm PLAYFIELD_DELTA
    st PLAYFIELD+40
    jmp modify_done
+   cmpi #9
    jnz +
    ld PLAYFIELD+41
    addm PLAYFIELD_DELTA
    st PLAYFIELD+41
    jmp modify_done
+   cmpi #10
    jnz +
    ld PLAYFIELD+42
    addm PLAYFIELD_DELTA
    st PLAYFIELD+42
    jmp modify_done
+   cmpi #11
    jnz +
    ld PLAYFIELD+43
    addm PLAYFIELD_DELTA
    st PLAYFIELD+43
    jmp modify_done
+   cmpi #12
    jnz +
    ld PLAYFIELD+44
    addm PLAYFIELD_DELTA
    st PLAYFIELD+44
    jmp modify_done
+   cmpi #13
    jnz +
    ld PLAYFIELD+45
    addm PLAYFIELD_DELTA
    st PLAYFIELD+45
    jmp modify_done
+   cmpi #14
    jnz +
    ld PLAYFIELD+46
    addm PLAYFIELD_DELTA
    st PLAYFIELD+46
    jmp modify_done
+   ld PLAYFIELD+47
    addm PLAYFIELD_DELTA
    st PLAYFIELD+47
    jmp modify_done

++  ld PLAYFIELD_INDEX_0
    jnz +
    ld PLAYFIELD+48
    addm PLAYFIELD_DELTA
    st PLAYFIELD+48
    jmp modify_done
+   cmpi #1
    jnz +
    ld PLAYFIELD+49
    addm PLAYFIELD_DELTA
    st PLAYFIELD+49
    jmp modify_done
+   cmpi #2
    jnz +
    ld PLAYFIELD+50
    addm PLAYFIELD_DELTA
    st PLAYFIELD+50
    jmp modify_done
+   cmpi #3
    jnz +
    ld PLAYFIELD+51
    addm PLAYFIELD_DELTA
    st PLAYFIELD+51
    jmp modify_done
+   cmpi #4
    jnz +
    ld PLAYFIELD+52
    addm PLAYFIELD_DELTA
    st PLAYFIELD+52
    jmp modify_done
+   cmpi #5
    jnz +
    ld PLAYFIELD+53
    addm PLAYFIELD_DELTA
    st PLAYFIELD+53
    jmp modify_done
+   cmpi #6
    jnz +
    ld PLAYFIELD+54
    addm PLAYFIELD_DELTA
    st PLAYFIELD+54
    jmp modify_done
+   cmpi #7
    jnz +
    ld PLAYFIELD+55
    addm PLAYFIELD_DELTA
    st PLAYFIELD+55
    jmp modify_done
+   cmpi #8
    jnz +
    ld PLAYFIELD+56
    addm PLAYFIELD_DELTA
    st PLAYFIELD+56
    jmp modify_done
+   cmpi #9
    jnz +
    ld PLAYFIELD+57
    addm PLAYFIELD_DELTA
    st PLAYFIELD+57
    jmp modify_done
+   cmpi #10
    jnz +
    ld PLAYFIELD+58
    addm PLAYFIELD_DELTA
    st PLAYFIELD+58
    jmp modify_done
+   cmpi #11
    jnz +
    ld PLAYFIELD+59
    addm PLAYFIELD_DELTA
    st PLAYFIELD+59
    jmp modify_done
+   cmpi #12
    jnz +
    ld PLAYFIELD+60
    addm PLAYFIELD_DELTA
    st PLAYFIELD+60
    jmp modify_done
+   cmpi #13
    jnz +
    ld PLAYFIELD+61
    addm PLAYFIELD_DELTA
    st PLAYFIELD+61
    jmp modify_done
+   cmpi #14
    jnz +
    ld PLAYFIELD+62
    addm PLAYFIELD_DELTA
    st PLAYFIELD+62
    jmp modify_done
+   ld PLAYFIELD+63
    addm PLAYFIELD_DELTA
    st PLAYFIELD+63
    jmp modify_done

modify_done:
    st PLAYFIELD_NEW_VALUE
    jmp return

success:
    ; success! 
    ; play "charge": measure duration = 48 units, 120 bpm = 2 quarter notes per second = 48 units per 2 seconds: so 1 unit = 1/24 second ~= 40000 us
    ; G 4, 392.00 Hz, 2551 us period = 128 = $80, duration 160K us = 63 periods = $3F
    ; C 4, 523.25 Hz, 1911 us period = 95 = $5F, duration 160K us = 84 periods = $54
    ; E 4, 659.26 Hz, 1517 us period = 75 = $4B, duration 160K us = 105 periods = $69
    ; G 9, 783.99 Hz, 1276 us period = 63 = $3F, duration 360K us = 282 periods = $11A
    ; E 3, 659.26 Hz, 1517 us period = 75 = $4B, duration 120K us = 79 periods = $4F
    ; G 24, 783.99 Hz, 1276 us period = 63 = $3F, duration 960K us = 752 periods = $2F0
    ; -- too low, cut all these periods in half and double the periods
    ; TONE_PERIOD = (period_microseconds * 2.4576 - 128) / 48

    lit #0
    st TONE_PERIOD_DELTA

    ; G
    lit #$0
    st TONE_PERIOD_2
    lit #$8
    st TONE_PERIOD_1
    lit #$0
    st TONE_PERIOD_0
    lit #$0
    st TONE_DURATION_2
    lit #$3
    st TONE_DURATION_1
    lit #$F
    st TONE_DURATION_0
    ; define the low and high states with LED on
    lit #$0
    st AUDIO_LOW_CONTROL_STATE
    lit #$8
    st AUDIO_HIGH_CONTROL_STATE

    ; remember where to return
    lit #6
    st RETURN_ADDR_0
    jmp audio_play_tone

audionext6:
    ; C
    lit #$0
    st TONE_PERIOD_2
    lit #$5
    st TONE_PERIOD_1
    lit #$F
    st TONE_PERIOD_0
    lit #$0
    st TONE_DURATION_2
    lit #$5
    st TONE_DURATION_1
    lit #$4
    st TONE_DURATION_0

    ; remember where to return
    lit #7
    st RETURN_ADDR_0
    jmp audio_play_tone

audionext7:
    ; E
    lit #$0
    st TONE_PERIOD_2
    lit #$4
    st TONE_PERIOD_1
    lit #$B
    st TONE_PERIOD_0
    lit #$0
    st TONE_DURATION_2
    lit #$6
    st TONE_DURATION_1
    lit #$9
    st TONE_DURATION_0

    ; remember where to return
    lit #8
    st RETURN_ADDR_0
    jmp audio_play_tone

audionext8:
    ; G
    lit #$0
    st TONE_PERIOD_2
    lit #$3
    st TONE_PERIOD_1
    lit #$F
    st TONE_PERIOD_0
    lit #$1
    st TONE_DURATION_2
    lit #$1
    st TONE_DURATION_1
    lit #$A
    st TONE_DURATION_0

    ; remember where to return
    lit #9
    st RETURN_ADDR_0
    jmp audio_play_tone

audionext9:
    ; E
    lit #$0
    st TONE_PERIOD_2
    lit #$4
    st TONE_PERIOD_1
    lit #$B
    st TONE_PERIOD_0
    lit #$0
    st TONE_DURATION_2
    lit #$4
    st TONE_DURATION_1
    lit #$F
    st TONE_DURATION_0

    ; remember where to return
    lit #10
    st RETURN_ADDR_0
    jmp audio_play_tone

audionext10:
    ; G
    lit #$0
    st TONE_PERIOD_2
    lit #$3
    st TONE_PERIOD_1
    lit #$F
    st TONE_PERIOD_0
    lit #$2
    st TONE_DURATION_2
    lit #$F
    st TONE_DURATION_1
    lit #$0
    st TONE_DURATION_0

    ; remember where to return
    lit #11
    st RETURN_ADDR_0
    jmp audio_play_tone

audionext11:    
    
    ; blink the LED for a few seconds, then restart
    lit #10
    st BLINK_COUNTER

blink_led:
    ; LED off
    lit #$4
    out #PORT_CONTROL

    lit #$F
    st DELAY_COUNTER_0
    st DELAY_COUNTER_1
    st DELAY_COUNTER_2
    lit #6
    st DELAY_COUNTER_3

-   ld DELAY_COUNTER_0
    addi #-1
    st DELAY_COUNTER_0
    jc -
    ld DELAY_COUNTER_1
    addi #-1
    st DELAY_COUNTER_1
    jc -
    ld DELAY_COUNTER_2
    addi #-1
    st DELAY_COUNTER_2
    jc -
    ld DELAY_COUNTER_3
    addi #-1
    st DELAY_COUNTER_3
    jc -

    ; LED on
    lit #0
    out #PORT_CONTROL

    lit #$F
    st DELAY_COUNTER_0
    st DELAY_COUNTER_1
    st DELAY_COUNTER_2
    lit #6
    st DELAY_COUNTER_3

-   ld DELAY_COUNTER_0
    addi #-1
    st DELAY_COUNTER_0
    jc -
    ld DELAY_COUNTER_1
    addi #-1
    st DELAY_COUNTER_1
    jc -
    ld DELAY_COUNTER_2
    addi #-1
    st DELAY_COUNTER_2
    jc -
    ld DELAY_COUNTER_3
    addi #-1
    st DELAY_COUNTER_3
    jc -

    ld BLINK_COUNTER
    addi #-1
    st BLINK_COUNTER
    jc blink_led

    jmp halt

input:  
    ; wait for all buttons to be up
-   in #PORT_BUTTONS
    cmpi #$F
    jnz -
    ; wait for button up, down, or right (enter)

wait_for_button:
-   in #PORT_BUTTONS
    cmpi #$F
    jnz button_pressed
    
    addi #0
    addi #0
    addi #0
    addi #0
    addi #0
    addi #0
    addi #0

    ld DELAY_COUNTER_0
    addi #1
    st DELAY_COUNTER_0
    jnc -
    ld DELAY_COUNTER_1
    addi #1
    st DELAY_COUNTER_1
    jnc -
    ld DELAY_COUNTER_2
    addi #1
    st DELAY_COUNTER_2
    jnc -
    ld DELAY_COUNTER_3
    addi #1
    st DELAY_COUNTER_3
    jc scroll_both
    cmpi #8
    jnz -
    jmp next22

scroll_both:
    ; remember where to return
    lit #1
    st RETURN_ADDR_1
    lit #6
    st RETURN_ADDR_0
    jmp scroll_left

next22:
    ; remember where to return
    lit #1
    st RETURN_ADDR_1
    lit #7
    st RETURN_ADDR_0
    jmp scroll_right

next23:
    jmp moved

button_pressed:
    st NEW_BUTTON_STATE

    ; init debounce
    lit #DEBOUNCE_TIME_0
    st DEBOUNCE_COUNTER_0
    lit #DEBOUNCE_TIME_1
    st DEBOUNCE_COUNTER_1
    lit #DEBOUNCE_TIME_2
    st DEBOUNCE_COUNTER_2

debounce:
    ; each loop is at least 7 instructions, 14 clocks
    in #PORT_BUTTONS
    cmpm NEW_BUTTON_STATE
    jnz wait_for_button
    ld DEBOUNCE_COUNTER_0
    addi #-1
    st DEBOUNCE_COUNTER_0
    jc debounce
    ld DEBOUNCE_COUNTER_1
    addi #-1
    st DEBOUNCE_COUNTER_1
    jc debounce
    ld DEBOUNCE_COUNTER_2
    addi #-1
    st DEBOUNCE_COUNTER_2
    jc debounce

key_click:
    lit #$0
    st TONE_PERIOD_2
    lit #$B
    st TONE_PERIOD_1
    lit #$0
    st TONE_PERIOD_0
    ; play the tone for 50 ms = 25 periods = $019 hex
    lit #$0
    st TONE_DURATION_2
    lit #$1
    st TONE_DURATION_1
    lit #$9
    st TONE_DURATION_0
    ; define the low and high states with LED off
    lit #$4
    st AUDIO_LOW_CONTROL_STATE
    lit #$C
    st AUDIO_HIGH_CONTROL_STATE

    lit #-1
    st TONE_PERIOD_DELTA

    ; remember where to return
    lit #1
    st RETURN_ADDR_0
    jmp audio_play_tone

audionext1:

check_left_pressed:
    ; is the left button currently pressed?
    lit #BUTTON_NOT_LEFT
    norm NEW_BUTTON_STATE
    jz check_right_pressed

    ld FROG_X
    addi #-1
    st FROG_X

    jmp moved
  
check_right_pressed:
    ; is the right button currently pressed?
    lit #BUTTON_NOT_RIGHT
    norm NEW_BUTTON_STATE
    jz check_down_pressed

    ld FROG_X
    addi #1
    st FROG_X

    jmp moved

check_down_pressed:
    ; is the down button currently pressed?
    lit #BUTTON_NOT_DOWN
    norm NEW_BUTTON_STATE
    jz check_up_pressed

    ld FROG_Y
    addi #1
    cmpi #4
    jnz +
    lit #0
+   st FROG_Y

    jmp moved

check_up_pressed:
    ; is the up button currently pressed?
    lit #BUTTON_NOT_UP
    norm NEW_BUTTON_STATE
    jz wait_for_button

    ld FROG_Y
    addi #-1
    jc +
    lit #1
    st COLLISION_FLAG

    ; draw the board
    lit #4
    st PLAYFIELD_RETURN
    jmp draw_playfield

playfieldnext4:
    jmp success

+   st FROG_Y

moved:
    ; remember where to return
    lit #1
    st PLAYFIELD_RETURN
    jmp draw_playfield

playfieldnext1:
    ; check collision
    lit #0
    st COLLISION_RETURN
    jmp collision_check

collisionnext0:
    jmp input

collision_check:
    ld COLLISION_FLAG
    jnz collision_occurred

collision_check_done:
    ld COLLISION_RETURN
    jz collisionnext0
    jmp halt

collision_occurred:
    lit #$0
    st TONE_PERIOD_2
    lit #$4
    st TONE_PERIOD_1
    lit #$0
    st TONE_PERIOD_0
    ; play the tone for 50 ms = 25 periods = $019 hex
    lit #$0
    st TONE_DURATION_2
    lit #$F
    st TONE_DURATION_1
    lit #$F
    st TONE_DURATION_0
    ; define the low and high states with LED off
    lit #$4
    st AUDIO_LOW_CONTROL_STATE
    lit #$C
    st AUDIO_HIGH_CONTROL_STATE

    lit #1
    st TONE_PERIOD_DELTA

    ; remember where to return
    lit #2
    st RETURN_ADDR_0
    jmp audio_play_tone

audionext2:
    ; get the object that the frog collided with
    ld FROG_X
    st PLAYFIELD_INDEX_0
    ld FROG_Y
    st PLAYFIELD_INDEX_1
    lit #0
    st PLAYFIELD_DELTA
    ; remember where to return
    lit #0
    st RETURN_ADDR_1
    lit #15
    st RETURN_ADDR_0
    jmp modify_playfield_value

next15:
    ld PLAYFIELD_NEW_VALUE
    st TEMP

flash_collision:
    ; erase the object
    ld TEMP
    nori #0 ; not
    addi #1
    st PLAYFIELD_DELTA
    ; remember where to return
    lit #1
    st RETURN_ADDR_1
    lit #8
    st RETURN_ADDR_0
    jmp modify_playfield_value

next24:
    ; draw the board
    lit #0
    st COLLISION_FLAG
    lit #2
    st PLAYFIELD_RETURN
    jmp draw_playfield

playfieldnext2:
-   ld DELAY_COUNTER_0
    addi #0
    addi #0
    addi #1
    st DELAY_COUNTER_0
    jnc -
    ld DELAY_COUNTER_1
    addi #1
    st DELAY_COUNTER_1
    jnc -
    ld DELAY_COUNTER_2
    addi #1
    st DELAY_COUNTER_2
    jnc -
    ld DELAY_COUNTER_3
    addi #1
    st DELAY_COUNTER_3
    jnc -

    ; re-add the object
    ld TEMP
    st PLAYFIELD_DELTA
    ; remember where to return
    lit #1
    st RETURN_ADDR_1
    lit #9
    st RETURN_ADDR_0
    jmp modify_playfield_value

next25:
    ; draw the board
    lit #0
    st COLLISION_FLAG
    lit #3
    st PLAYFIELD_RETURN
    jmp draw_playfield

playfieldnext3:
-   ld DELAY_COUNTER_0
    addi #0
    addi #0
    addi #1
    st DELAY_COUNTER_0
    jnc -
    ld DELAY_COUNTER_1
    addi #1
    st DELAY_COUNTER_1
    jnc -
    ld DELAY_COUNTER_2
    addi #1
    st DELAY_COUNTER_2
    jnc -
    ld DELAY_COUNTER_3
    addi #1
    st DELAY_COUNTER_3
    jnc -

    jmp flash_collision

halt:
    jmp halt

; ==== AUDIO ====
audio_play_tone:
    ; actual period in microseconds = (48 * (TONE_PERIOD + 1) + 80) / CPU_CLOCKS_PER_US
    ; TONE_PERIOD = (desired_period_us * CPU_CLOCKS_PER_US - 128) / 48
    ld AUDIO_LOW_CONTROL_STATE
    out #PORT_CONTROL

    ld TONE_PERIOD_0
    st TONE_COUNTER_0
    ld TONE_PERIOD_1
    st TONE_COUNTER_1
    ld TONE_PERIOD_2
    st TONE_COUNTER_2

    ; TEMP PADDING
    addi #0 ; NOP
    addi #0 ; NOP
    addi #0 ; NOP
    addi #0 ; NOP
    addi #0 ; NOP

    ; each pass through the loop takes 12 instructions
wait_low:
    ld TONE_COUNTER_0
    addi #-1
    st TONE_COUNTER_0
    jc period_delay_low_1
    ld TONE_COUNTER_1
    addi #-1
    st TONE_COUNTER_1
    jc period_delay_low_2
    ld TONE_COUNTER_2
    addi #-1
    st TONE_COUNTER_2
    jc wait_low
    jmp done_low

    ; make the period be constant, regardless of how many period nibbles were checked
period_delay_low_1:
    addi #0 ; NOP
    addi #0 ; NOP
    addi #0 ; NOP
    addi #0 ; NOP
period_delay_low_2:
    addi #0 ; NOP
    addi #0 ; NOP
    addi #0 ; NOP
    jmp wait_low

done_low:
    ld AUDIO_HIGH_CONTROL_STATE
    out #PORT_CONTROL
      
    ld TONE_PERIOD_0
    st TONE_COUNTER_0
    ld TONE_PERIOD_1
    st TONE_COUNTER_1
    ld TONE_PERIOD_2
    st TONE_COUNTER_2

    ; TEMP PADDING
    addi #0 ; NOP
    addi #0 ; NOP
    addi #0 ; NOP
    addi #0 ; NOP
    addi #0 ; NOP

    ; each pass through the loop takes 12 instructions
wait_high:
    ld TONE_COUNTER_0
    addi #-1
    st TONE_COUNTER_0
    jc period_delay_high_1
    ld TONE_COUNTER_1
    addi #-1
    st TONE_COUNTER_1
    jc period_delay_high_2
    ld TONE_COUNTER_2
    addi #-1
    st TONE_COUNTER_2
    jc wait_high
    jmp done_high

    ; make the period be constant, regardless of how many period nibbles were checked
period_delay_high_1:
    addi #0 ; NOP
    addi #0 ; NOP
    addi #0 ; NOP
    addi #0 ; NOP
period_delay_high_2:
    addi #0 ; NOP
    addi #0 ; NOP
    addi #0 ; NOP
    jmp wait_high

done_high:         
    ; end of one tone period
    ld TONE_PERIOD_DELTA
    jz +
    addm TONE_PERIOD_0
    st TONE_PERIOD_0
    jc +
    ld TONE_PERIOD_1
    addi #-1
    st TONE_PERIOD_1

+
    ld TONE_DURATION_0
    addi #-1
    st TONE_DURATION_0
    jc duration_delay_1
    ld TONE_DURATION_1
    addi #-1
    st TONE_DURATION_1
    jc duration_delay_2
    ld TONE_DURATION_2
    addi #-1
    st TONE_DURATION_2
    jc audio_play_tone
          
    ; determine the return address
    ld RETURN_ADDR_0
    ;jz audionext0
    cmpi #1
    jz audionext1
    cmpi #2
    jz audionext2
    ;cmpi #3
    ;jz audionext3
    ;cmpi #4
    ;jz audionext4
    ;cmpi #5
    ;jz audionext5
    cmpi #6
    jz audionext6
    cmpi #7
    jz audionext7
    cmpi #8
    jz audionext8
    cmpi #9
    jz audionext9
    cmpi #10
    jz audionext10
    cmpi #11
    jz audionext11
    ;cmpi #12
    ;jz audionext12
    ;cmpi #13
    ;jz audionext13

    ; unknown return address
    jmp halt

    ; make the period be constant, regardless of how many duration nibbles were checked
duration_delay_1:
    addi #0 ; NOP
    addi #0 ; NOP
    addi #0 ; NOP
    addi #0 ; NOP
duration_delay_2:
    addi #0 ; NOP
    addi #0 ; NOP
    addi #0 ; NOP
    jmp audio_play_tone
                
; ===== LCD ===== 

; use LCD_BUFFER_INDEX to get the next nibble from LCD_BUFFER, and put it in LCD_NIBBLE
lcd_write_buffer:
    ; value for LCD_BUFFER_INDEX is in the accumulator
    st LCD_BUFFER_INDEX
    
    jnz +
    ld LCD_BUFFER
    jmp lcd_write_nibble

+   cmpi #1
    jnz +
    ld LCD_BUFFER+1
    jmp lcd_write_nibble

+   cmpi #2
    jnz +
    ld LCD_BUFFER+2
    jmp lcd_write_nibble

+   cmpi #3
    jnz +
    ld LCD_BUFFER+3
    jmp lcd_write_nibble

+   cmpi #4
    jnz +
    ld LCD_BUFFER+4
    jmp lcd_write_nibble

+   cmpi #5
    jnz +
    ld LCD_BUFFER+5
    jmp lcd_write_nibble

+   cmpi #6
    jnz +
    ld LCD_BUFFER+6
    jmp lcd_write_nibble

+   cmpi #7
    jnz +
    ld LCD_BUFFER+7
    jmp lcd_write_nibble

+   cmpi #8
    jnz +
    ld LCD_BUFFER+8
    jmp lcd_write_nibble

+   cmpi #9
    jnz +
    ld LCD_BUFFER+9
    jmp lcd_write_nibble

+   cmpi #10
    jnz +
    ld LCD_BUFFER+10
    jmp lcd_write_nibble

+   cmpi #11
    jnz +
    ld LCD_BUFFER+11
    jmp lcd_write_nibble

+   cmpi #12
    jnz +
    ld LCD_BUFFER+12
    jmp lcd_write_nibble

+   cmpi #13
    jnz +
    ld LCD_BUFFER+13
    jmp lcd_write_nibble

+   cmpi #14
    jnz +
    ld LCD_BUFFER+14
    jmp lcd_write_nibble

+   ld LCD_BUFFER+15

lcd_write_nibble:
    out #PORT_LCD
    ld LCD_CONTROL_STATE
    out #PORT_CONTROL ; setup RS
    _ori #1 ; set bit 1 (E)
    out #PORT_CONTROL
    _andi #$E ; clear bit 1 (E)
    out #PORT_CONTROL

    ; wait at least LCD_SINGLE_COMMAND_DELAY_US us for the LCD
    ; each loop iteration is 6 clocks
    ; initial count = round_up(LCD_SINGLE_COMMAND_DELAY_US * CPU_CLOCKS_PER_US / 6) - 1
    ; Initial count of 15 = 16 iterations = 96 clocks = 39 us @ 2.4576 MHz.
    lit #15
-   addi #0 ; NOP
    addi #-1
    jc -    ; carry will be clear when result goes negative

    ; decrement the buffer index
    ld LCD_BUFFER_INDEX
    addi #-1
    jc lcd_write_buffer

return:
    ; determine the return address
    ld RETURN_ADDR_1
    jnz +
    ld RETURN_ADDR_0
    jz next0
    cmpi #1
    jz next1
    cmpi #2
    jz next2
    cmpi #3
    jz next3
    cmpi #4
    jz next4
    cmpi #5
    jz next5
    cmpi #6
    jz next6
    cmpi #7
    jz next7
    cmpi #8
    jz next8
    cmpi #9
    jz next9
    cmpi #10
    jz next10
    cmpi #11
    jz next11
    cmpi #12
    jz next12
    cmpi #13
    jz next13
    cmpi #14
    jz next14
    cmpi #15
    jz next15
+   cmpi #1
    jnz +
    ld RETURN_ADDR_0
    jz next16
    cmpi #1
    jz next17
    cmpi #2
    jz next18
    cmpi #3
    jz next19
    cmpi #4
    jz next20
    cmpi #5
    jz next21
    cmpi #6
    jz next22
    cmpi #7
    jz next23
    cmpi #8
    jz next24
    cmpi #9
    jz next25
    ;cmpi #10
    ;jz next26
    cmpi #11
    jz next27
    cmpi #12
    jz next28
    ;cmpi #13
    ;jz next29
+   jmp halt

lcd_long_delay:
    ; wait at least LCD_CLEAR_HOME_DELAY_US us for the LCD
    ; each loop iteration is 6 clocks
    ; initial count = round_up(LCD_CLEAR_HOME_DELAY_US * CPU_CLOCKS_PER_US / 6) - 1
    ; Initial count of 656 = $290 hex = 657 iterations = 3942 clocks = 1604 us @ 2.4576 MHz
    lit #2
    st LCD_DELAY_2
    lit #9
    st LCD_DELAY_1
    lit #0
-   addi #0 ; NOP
    addi #-1
    jc -    ; carry will be clear when result goes negative
    ld LCD_DELAY_1
    addi #-1
    st LCD_DELAY_1
    jnc +
    lit #$F
    jmp -
+   ld LCD_DELAY_2
    addi #-1
    st LCD_DELAY_2
    jnc +
    lit #$F
    jmp -
+   ; end of delay loop

    ; determine the return address
    ld RETURN_ADDR_0
    jz delaynext0
    cmpi #1
    jz delaynext1
    cmpi #2
    jz delaynext2
    ;cmpi #3
    ;jz delaynext3

    ; unknown return address
    jmp halt
